from utilities.forms import BootstrapMixin, StaticSelect2
from extras.forms import CustomFieldModelForm
from .models import Settings


class SettingsForm(BootstrapMixin, CustomFieldModelForm):
    class Meta:
        model = Settings
        fields = [
            "hostname",
            "username",
            "password",
            "version",
            "verify",
            "status",
        ]
        widgets = {
            "status": StaticSelect2(
                choices=(
                    ("True", "Yes"),
                    ("False", "No"),
                )
            ),
            "verify": StaticSelect2(
                choices=(
                    (True, "Yes"),
                    (False, "No"),
                )
            ),
        }



import re
from django_rq import get_worker
from django_rq.queues import get_connection
from extras.models import Tag
from dcim.models import Site
from tenancy.models import Tenant

class Plugin:
    name = "ciscodnacnetbox"
    version = "1.0"
    description = "Cisco DNA Center Integration with NetBox"
    author = "Robert Csapo"
    author_email = "rcsapo@cisco.com"

class System:
    """
    Support functions for the Plugin
    """

    class Check:
        @classmethod
        def tenant(cls, tenant):
            return Tenant.objects.filter(name=tenant).exists()

        @classmethod
        def sites(cls, tenant):
            if cls.tenant(tenant):
                t = Tenant.objects.get(name=tenant).id
                if Site.objects.filter(tenant=t).exists():
                    return True
            return False

    class PluginTag:
        @staticmethod
        def get():
            return Tag.objects.get(slug="cisco-dna-center")

        @staticmethod
        def filter():
            return Tag.objects.filter(slug="cisco-dna-center")

    class Slug:
        def create(input):
            return re.sub(r"[\s\/]+", "-", input).lower()

    class RQ:
        @staticmethod
        def status():
            if get_worker("default").count(get_connection("default")) == 0:
                return False
            return True



from . import CiscoDNAC
from cacheops import cache, CacheMiss
from dcim.models import Site, Device
from dcim.choices import DeviceStatusChoices
from tenancy.models import Tenant
from django_rq import get_queue, job
from ..models import Settings
from .netbox import Netbox
from .utilities import System


@job("default")
def full_sync(**kwargs):
    """
    RQ Background Task for Syncing Cisco DNA Center Instances
    """
    data = {}

    # Sync all func from Cisco DNA Center
    sites = Data.sync_sites(**kwargs)
    devices = Data.sync_devices(**kwargs)

    # Count the synced items
    for tenant in sites:
        data[tenant] = {}
        Netbox.Purge.database(tenant=tenant, type="devices", data=devices[tenant])
        Netbox.Purge.database(tenant=tenant, type="sites", data=sites[tenant])
    for tenant in sites:
        data[tenant]["sites"] = len(sites[tenant])
    for tenant in devices:
        data[tenant]["devices"] = len(devices[tenant])

    # Return data as results for the job
    return data


class Data:
    def status():
        """
        Plugin Status Dashboard
        """

        # Create Cisco DNA Center API Instances
        data = {}
        tenants = CiscoDNAC()
        data["dnac"] = {}

        # Get local settings
        for tenant in Settings.objects.all().nocache():
            data["dnac"][tenant.hostname] = {}
            data["dnac"][tenant.hostname]["id"] = tenant.id
            data["dnac"][tenant.hostname]["sites"] = None
            data["dnac"][tenant.hostname]["devices"] = None
        # Get API status per Cisco DNA Center

        for k, v in tenants.dnac_status.items():
            data["dnac"][k]["api"] = v

        # Get data from Cisco DNA Center
        for tenant, dnac in tenants.dnac.items():
            data["dnac"][tenant]["sites"] = tenants.sites_count(tenant=dnac)
            data["dnac"][tenant]["devices"] = 0
            for device in tenants.devices(tenant=dnac):
                if device.deviceSupportLevel == "Supported":
                    data["dnac"][tenant]["devices"] += 1

        # Gather data from NetBox
        data["netbox"] = {}
        dnac_tag = System.PluginTag.get()
        data["netbox"]["sites"] = len(Site.objects.filter(tags=dnac_tag))
        data["netbox"]["devices"] = len(Device.objects.filter(tags=dnac_tag))
        data["netbox"]["tenants"] = {}

        # Gather Tenants that is related to Cisco DNA Center
        for tenant in Tenant.objects.filter(tags=dnac_tag):
            managed = False
            if str(tenant) in [*data["dnac"]]:
                managed = True
            data["netbox"]["tenants"][tenant.name] = {
                "id": tenant.id,
                "description": tenant.description,
                "created": tenant.created,
                "managed": managed,
            }
        return data

    def devices(**kwargs):
        """
        Cisco DNA Center Instance Devices
        """

        # Gather all devices from Cisco DNA Center Inventory (no sync)
        data = {}
        tenants = CiscoDNAC(**kwargs)
        for tenant, dnac in tenants.dnac.items():
            data[tenant] = tenants.devices(tenant=dnac)
        return data

    def sites(**kwargs):
        """
        Cisco DNA Center Sites Devices
        """

        # Gather all sites from Cisco DNA Center Network Designs (no sync)
        data = {}
        tenants = CiscoDNAC(**kwargs)
        for tenant, dnac in tenants.dnac.items():
            results = []
            for site in tenants.sites(tenant=dnac):

                # Get addtional data about the location
                if len(site.additionalInfo) != 0:
                    for additionalInfo in site.additionalInfo:
                        if "Location" in additionalInfo["nameSpace"]:
                            result = {
                                "name": site.name,
                                "siteNameHierarchy": site.siteNameHierarchy,
                                "type": additionalInfo["attributes"]["type"],
                                "country": additionalInfo["attributes"]["country"],
                            }
                else:
                    result = {
                        "name": site.name,
                        "siteNameHierarchy": site.siteNameHierarchy,
                        "type": None,
                        "country": None,
                    }
                results.append(result)
            results = sorted(results, key=lambda k: k["siteNameHierarchy"], reverse=False)
            data[tenant] = results
        return data

    @classmethod
    def sync_full(cls, **kwargs):
        """
        Sync Cisco DNA Center as RQ job
        """
        data = {}

        # Get RQ queue
        queue = get_queue("default")

        # Get RQ Job ID and display results
        if "id" in kwargs:
            data = queue.fetch_job(str(kwargs["id"]))
            if data is None:
                return None
            return data.result

        # Check if ongoing RQ Job is ongoing
        try:
            job = cache.get("ciscodnacnetbox_bg")
        except CacheMiss:
            # If not, start full sync task
            job = full_sync.delay(**kwargs)
            cache.set("ciscodnacnetbox_bg", job.id, timeout=600)

        # Get Job Status
        j = queue.fetch_job(cache.get("ciscodnacnetbox_bg"))
        if "finished" == j.get_status():
            # Start again, if cache expired
            job = full_sync.delay(**kwargs)
            cache.set("ciscodnacnetbox_bg", job.id, timeout=600)
        data["id"] = str(j.id)
        data["task"] = str(j.func_name)
        return data

    @classmethod
    def sync_sites(cls, **kwargs):
        """
        Sync Cisco DNA Center Sites
        """

        # Sync mandatory tag for Cisco DNA Center in NetBox
        dnac_tag = Netbox.Sync.tags(task="system")

        # Gather all sites in Cisco DNA Center Network Designs
        data = {}
        tenants = CiscoDNAC(**kwargs)
        for tenant, dnac in tenants.dnac.items():
            results = []
            # Sync Cisco DNA Center Tenant
            Netbox.Sync.tenants(task="system", tenant=tenant, slug=tenant.replace(".", "-"))
            # Add tag to Cisco DNA Center Tenant
            Netbox.Sync.tags(
                task="update",
                model="tenant",
                filter=tenant,
                tag=dnac_tag,
            )
            for site in tenants.sites(tenant=dnac):
                # Sync Site
                # Unique name for `Global` as it can't be duplicate in NetBox
                if site.siteNameHierarchy == "Global":
                    suffix = site.id.split("-")
                    site.siteNameHierarchy = "{} {}".format(site.siteNameHierarchy, suffix[0])

                # Use Cisco DNA Center UUID for Site as Slug
                site.slug = site.id
                site.sync = Netbox.Sync.site(tenant=tenant, site=site)

                # Add tag to Site
                Netbox.Sync.tags(
                    task="update",
                    model="site",
                    filter=site.siteNameHierarchy,
                    tag=dnac_tag,
                )

                site.status = "Active"
                site.status_label = "success"
                result = {
                    "name": site.name,
                    "status": site.status,
                    "status_label": site.status_label,
                    "slug": site.slug,
                    "sync_status": site.sync[1],
                }
                results.append(result)

            # If site is removed in Cisco DNA Center, then remove in NetBox
            Netbox.Purge.database(tenant=tenant, type="sites", data=results)
            results = sorted(results, key=lambda k: k["name"], reverse=False)
            data[tenant] = results
        return data

    @classmethod
    def sync_devices(cls, **kwargs):
        """
        Sync Cisco DNA Center Devices
        """

        # Sync mandatory tag for Cisco DNA Center
        dnac_tag = Netbox.Sync.tags(task="system")

        # Gather all devices in Cisco DNA Center Inventory
        data = {}
        tenants = CiscoDNAC(**kwargs)
        for tenant, dnac in tenants.dnac.items():
            results = []

            # NetBox sites mandatory to assign sites
            if System.Check.sites(tenant=tenant) is False:
                data[tenant] = [{"sync_status": "Error: Sync sites first"}]
                continue
            # Map Devices (Serial) against Site UUID
            site_members = CiscoDNAC.devices_to_sites(tenant=dnac)

            # Get devices from Cisco DNA Center
            for device in tenants.devices(tenant=dnac):

                # Sync Cisco DNA Center Tenant
                Netbox.Sync.tenants(task="system", tenant=tenant, slug=tenant.replace(".", "-"))
                Netbox.Sync.tags(
                    task="update",
                    model="tenant",
                    filter=tenant,
                    tag=dnac_tag,
                )

                # Check that the device is supported in Cisco DNA Center
                if device.deviceSupportLevel == "Supported":

                    # Sync Manufacture
                    device.manufacture = device.type.split()[0]
                    device.manufacture = Netbox.Sync.manufacturer(manufacture=device.manufacture, tenant=tenant)

                    # Sync Device Types
                    slug = System.Slug.create(device.family)
                    device.family_type = Netbox.Sync.devicetype(
                        manufacture=device.manufacture,
                        model=device.family,
                        slug=slug,
                        tenant=tenant,
                    )
                    # Add tag to devicetype
                    Netbox.Sync.tags(
                        task="update",
                        model="devicetype",
                        filter=slug,
                        tag=dnac_tag,
                    )

                    # Sync Device Roles
                    slug = System.Slug.create(device.role)
                    device.role = Netbox.Sync.devicerole(role=device.role, slug=slug, tenant=tenant)

                    # Sync Device IP Address
                    device.primary_ip4 = Netbox.Sync.ipaddress(
                        tenant=tenant,
                        address=device.managementIpAddress,
                        hostname=device.hostname,
                    )
                    # Add tags to IP Address
                    Netbox.Sync.tags(
                        task="update",
                        model="ipaddress",
                        filter=device.primary_ip4,
                        tag=dnac_tag,
                    )
                    # Device Site Location
                    device.site = Site.objects.get(
                        slug=site_members[device.serialNumber],
                        tenant=Tenant.objects.get(name=tenant).id,
                    )

                    # Check if devices is reachable from Cisco DNA Center
                    if device.reachabilityStatus == "Reachable":
                        device.status = DeviceStatusChoices.STATUS_ACTIVE
                        device.status_label = "success"
                    else:
                        device.status = DeviceStatusChoices.STATUS_FAILED
                        device.status_label = "danger"

                    # Sync Device and get status
                    sync_status = Netbox.Sync.device(tenant=tenant, device=device)
                    # Add tag to device
                    Netbox.Sync.tags(
                        task="update",
                        model="device",
                        filter=device.serialNumber,
                        tag=dnac_tag,
                    )
                    result = {
                        "name": device.hostname,
                        "status": device.status,
                        "status_label": device.status_label,
                        "role": device.role,
                        "type": device.family_type,
                        "site": device.site,
                        "primary_ip4": device.primary_ip4,
                        "serial": device.serialNumber,
                        "sync_status": sync_status[1],
                    }
                    results.append(result)

            # If device is removed in Cisco DNA Center, then remove in NetBox
            Netbox.Purge.database(tenant=tenant, type="devices", data=results)

            results = sorted(results, key=lambda k: k["name"], reverse=False)
            data[tenant] = results
        return data

    def purge_tenant(**kwargs):
        """
        Remove NetBox Tenant that is related to Cisco DNA Center
        """
        results = Netbox.Purge.tenant(**kwargs)
        return results

    @staticmethod
    def job_status(id):
        """
        Get RQ Job Status
        """

        # Get RQ queue
        data = {}
        queue = get_queue("default")
        # Get Job Id data
        j = queue.fetch_job(str(id))
        if j is None:
            # No job exists with that `id`
            return None
        data["id"] = str(id)
        data["task"] = str(j.func_name)
        data["status"] = str(j.get_status())
        data["result"] = str(j.result)
        data["exception"] = str(j.exc_info)
        return data



from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True
    dependencies = []
    operations = [
        migrations.CreateModel(
            name="Settings",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False),
                ),
                ("hostname", models.CharField(max_length=2000, unique=True)),
                ("username", models.CharField(max_length=100)),
                ("password", models.CharField(max_length=100)),
                ("version", models.CharField(max_length=10)),
                ("verify", models.BooleanField(default=False)),
                ("status", models.BooleanField(default=True)),
            ],
            options={
                "ordering": ["hostname"],
            },
        ),
    ]



import platform
from django.conf import settings
from django.http import Http404, HttpResponseServerError, JsonResponse
from django.views.defaults import ERROR_500_TEMPLATE_NAME
from django.template import loader
from django.urls import reverse
from django.shortcuts import get_object_or_404, render, redirect
from django.views.generic import View
from utilities.forms import ConfirmationForm
from tenancy.models import Tenant
from netbox.views import generic
from .models import Settings
from .forms import SettingsForm
from .tables import SettingsTable
from .ciscodnac.data import Data
from .ciscodnac.netbox import Netbox
from .ciscodnac.utilities import System


class SettingsView(generic.ObjectListView):
    """
    Cisco DNA Center Settings
    """
    
    queryset = Settings.objects.all()
    table = SettingsTable
    template_name = "ciscodnacnetbox/settings.html"


class SettingsEdit(generic.ObjectEditView):
    """
    Add/Edit Cisco DNA Center Settings
    """

    queryset = Settings.objects.all()
    model_form = SettingsForm
    template_name = "ciscodnacnetbox/settings_edit.html"


class SettingsDelete(generic.ObjectDeleteView):
    """
    Delete Cisco DNA Center Settings
    """

    queryset = Settings.objects.all()


class SettingsDeleteBulk(generic.BulkDeleteView):
    """
    Delete multiple Cisco DNA Center Settings
    """

    queryset = Settings.objects.all()
    table = SettingsTable


class StatusView(View):
    """
    Plugin Status Dashboard
    """

    def get(self, request):
        # Check that NetBox tag exists for Cisco DNA Center
        Netbox.Sync.tags(task="system")

        # Check that Cisco DNA Center Settings exists
        if Settings.objects.filter().exists() is False:
            return redirect("/plugins/ciscodnacnetbox/settings/")

        data = Data.status()
        return render(
            request,
            "ciscodnacnetbox/status.html",
            {
                "dnac": data["dnac"],
                "netbox": request.build_absolute_uri("/"),
                "netbox_sites": data["netbox"]["sites"],
                "netbox_devices": data["netbox"]["devices"],
                "netbox_tenants": data["netbox"]["tenants"],
            },
        )


class SyncFull(View):
    """
    Sync Cisco DNA Center
    """

    def get(self, request, **kwargs):
        # Check that we have Cisco DNA Center settings
        if Settings.objects.filter().exists() is False:
            return redirect("/plugins/ciscodnacnetbox/settings/")

        # Check if RQ workers are running
        if System.RQ.status() is False:
            template = loader.get_template(ERROR_500_TEMPLATE_NAME)
            error_msg = """
            Addtional Workers not running for Background Tasks.
            Verify that rqworker is running.
            """
            return HttpResponseServerError(
                template.render(
                    {
                        "error": error_msg,
                        "exception": "ciscodnacnetbox plugin - RQ",
                        "netbox_version": settings.VERSION,
                        "python_version": platform.python_version(),
                    }
                )
            )

        # Run Sync as Background Job in RQ
        data = Data.sync_full(**kwargs)
        if "id" in kwargs:
            if data is None:
                raise Http404()
            return render(
                request,
                "ciscodnacnetbox/sync_full.html",
                {
                    "data": data,
                },
            )
        return render(
            request,
            "ciscodnacnetbox/loading_job.html",
            {
                "data": data,
            },
        )


class SyncFullFailed(View):
    """
    Display failed RQ Job
    """

    def get(self, request, id):
        data = Data.job_status(id)
        return render(
            request,
            "ciscodnacnetbox/sync_full_failed.html",
            {
                "data": data,
            },
        )


class JobStatus(View):
    """
    Check RQ Job Status
    """

    def get(self, request, id):
        data = Data.job_status(id)
        if data is None:
            raise Http404()
        return JsonResponse(data)


class DeviceView(View):
    """
    Cisco DNA Center Devices
    """

    def get(self, request, **kwargs):
        data = Data.devices(**kwargs)
        return render(
            request,
            "ciscodnacnetbox/devices.html",
            {
                "data": data,
            },
        )


class SyncDevices(View):
    """
    Sync Cisco DNA Center Devices
    """

    def get(self, request, **kwargs):
        data = Data.sync_devices(**kwargs)
        return render(
            request,
            "ciscodnacnetbox/sync_devices.html",
            {
                "data": data,
            },
        )


class SitesView(View):
    """
    Cisco DNA Center Sites
    """

    def get(self, request, **kwargs):
        data = Data.sites(**kwargs)
        return render(
            request,
            "ciscodnacnetbox/sites.html",
            {
                "data": data,
            },
        )


class SyncSites(View):
    """
    Sync Cisco DNA Center Sites
    """

    def get(self, request, **kwargs):
        data = Data.sync_sites(**kwargs)
        return render(
            request,
            "ciscodnacnetbox/sync_sites.html",
            {
                "data": data,
            },
        )


class PurgeTenant(View):
    """
    Purge NetBox Tenants related to Cisco DNA Center
    """

    def get(self, request, **kwargs):

        # Verify that the Tenant exists in NetBox
        tenant = get_object_or_404(Tenant, pk=kwargs["pk"], tags=System.PluginTag.get())

        # Confirm deletion
        form = ConfirmationForm(initial=request.GET)
        return render(
            request,
            "generic/object_delete.html",
            {
                "obj": tenant,
                "form": form,
                "return_url": reverse("plugins:ciscodnacnetbox:purge_tenant", args=(kwargs["pk"],)),
            },
        )

    def post(self, request, **kwargs):

        # Delete Tenant in NetBox
        data = Data.purge_tenant(**kwargs)
        return render(
            request,
            "ciscodnacnetbox/purge_tenant.html",
            {
                "data": data,
            },
        )



from extras.plugins import PluginConfig

class CiscoDNACenterConfig(PluginConfig):
    version = "1.0"
    name = "ciscodnacnetbox"
    verbose_name = "Cisco DNA Center Sync Plugin"
    description = "Cisco DNA Center Integration with NetBox"
    author = "Robert Csapo"
    author_email = "rcsapo@cisco.com"
    required_settings = []
    default_settings = {}
    base_url = "ciscodnacnetbox"
    caching_config = {}


config = CiscoDNACenterConfig



from django.shortcuts import get_object_or_404
from dnacentersdk import api
from ..models import Settings


class CiscoDNAC:

    # Get all tenants from Settings
    __tenants = Settings.objects.all().nocache()

    def __init__(self, **kwargs):
        """
        Cisco DNA Center API Instance
        """
        self.dnac = {}
        self.dnac_status = {}

        # Single Cisco DNA Center Instance
        if "pk" in kwargs and isinstance(kwargs["pk"], int) is True:

            # Verify that the Tenant exist based on the `pk`
            tenant = get_object_or_404(Settings, pk=kwargs["pk"])

            # Create Cisco DNA Center API Object
            obj = self.auth(tenant)

            # Check that Auth is successful
            if obj:
                self.dnac[tenant.hostname] = obj[1]
            return

        for tenant in self.__tenants:
            self.dnac_status[tenant.hostname] = "disabled"

            # Create Cisco DNA Center API Object if enabled
            if tenant.status is True:

                # Create Cisco DNA Center API Object
                obj = self.auth(tenant)

                # Check that Auth is successful
                if obj:
                    self.dnac[tenant.hostname] = obj[1]
        return

    def auth(self, tenant):
        """
        Cisco DNA Center API Object
        """
        try:
            obj = api.DNACenterAPI(
                username=tenant.username,
                password=tenant.password,
                base_url="https://" + tenant.hostname,
                # version="2.1.2",  # TODO
                verify=bool(tenant.verify),
            )
            self.dnac_status[tenant.hostname] = "success"
            return True, obj
        except Exception as error_msg:
            print("Error for {}: {}".format(tenant, error_msg))
            self.dnac_status[tenant.hostname] = error_msg
            return False

    def devices(self, tenant):
        """
        Get Devices from Cisco DNA Center
        """
        return tenant.devices.get_device_list().response

    def sites(self, tenant):
        """
        Get Sites from Cisco DNA Center
        """
        return tenant.sites.get_site().response

    def sites_count(self, tenant):
        """
        Get Sites count from Cisco DNA Center
        """
        return tenant.sites.get_site_count().response

    @classmethod
    def devices_to_sites(cls, tenant):
        """
        Map Device Serial Number to Site ID from Cisco DNA Center
        """
        results = {}
        for site in tenant.sites.get_site().response:
            for members in tenant.sites.get_membership(site_id=site.id).device:
                for device in members.response:
                    results[device.serialNumber] = site.id
        return results



import django_tables2 as tables
from django_tables2.utils import A
from django.utils.safestring import mark_safe
from utilities.tables import (
    BaseTable,
    BooleanColumn,
    ToggleColumn,
)
from .models import Settings


class MaskedPassword(tables.Column):
    def render(self, value):
        value = "*****"
        return mark_safe(value)


class SettingsTable(BaseTable):
    pk = ToggleColumn()
    hostname = tables.LinkColumn(
        "plugins:ciscodnacnetbox:settings_edit", args=[A("pk")]
    )
    username = tables.Column()
    password = MaskedPassword()
    version = tables.Column()
    verify = BooleanColumn()
    status = BooleanColumn()

    class Meta(BaseTable.Meta):
        model = Settings
        fields = [
            "pk",
            "hostname",
            "username",
            "password",
            "version",
            "verify",
            "status",
        ]



from extras.plugins import PluginMenuButton, PluginMenuItem
from utilities.choices import ButtonColorChoices


menu_items = (
    PluginMenuItem(
        link="plugins:ciscodnacnetbox:status",
        link_text="Status",
        permissions=["admin"],
        buttons=(
            PluginMenuButton(
                link="plugins:ciscodnacnetbox:sync_full",
                title="Settings",
                icon_class="mdi mdi-all-inclusive",
                color=ButtonColorChoices.BLUE,
                permissions=["admin"],
            ),
            PluginMenuButton(
                link="plugins:ciscodnacnetbox:settings",
                title="Settings",
                icon_class="mdi mdi-cog",
                color=ButtonColorChoices.BLUE,
                permissions=["admin"],
            ),
        ),
    ),
)



from decimal import Decimal
import ipaddress
from django.shortcuts import get_object_or_404
from extras.models import Tag
from dcim.models import Site, Device, DeviceRole, DeviceType, Manufacturer
from ipam.models import IPAddress
from dcim.choices import DeviceStatusChoices
from tenancy.models import Tenant
from utilities.choices import ColorChoices
from .utilities import System


class Netbox:
    class Sync:
        """
        Sync data to NetBox Models
        """

        @staticmethod
        def tenants(**kwargs):
            """
            Create Tenant based on Cisco DNA Center Instance
            """
            if "system" in kwargs["task"]:
                if Tenant.objects.filter(name=kwargs["tenant"]).exists() is False:
                    Tenant.objects.create(
                        name=kwargs["tenant"],
                        slug=kwargs["slug"],
                        description="Managed by {}".format(
                            kwargs["tenant"],
                        ),
                    )
                else:
                    Tenant.objects.filter(name=kwargs["tenant"]).update(
                        description="Managed by {}".format(
                            kwargs["tenant"],
                        ),
                    )
                return Tenant.objects.get(name=kwargs["tenant"])

        @staticmethod
        def tags(**kwargs):
            """
            Handle Tag operations with NetBox
            """
            if "system" in kwargs["task"]:
                # Create mandatory Cisco DNA Center Tag
                if len(System.PluginTag.filter()) == 0:
                    Tag.objects.create(
                        name="Cisco DNA Center",
                        slug="cisco-dna-center",
                        color=ColorChoices.COLOR_BLUE,
                        description="Managed by ciscodnacnetbox",
                    )
                else:
                    System.PluginTag.filter().update(
                        name="Cisco DNA Center",
                        description="Managed by ciscodnacnetbox",
                    )
                return System.PluginTag.get()
            elif "update" in kwargs["task"]:
                # Get Object before saving Tag
                if kwargs["model"].lower() == "Tenant".lower():
                    __obj = Tenant.objects.get(name=kwargs["filter"])
                if kwargs["model"].lower() == "DeviceType".lower():
                    __obj = DeviceType.objects.get(slug=kwargs["filter"])
                if kwargs["model"].lower() == "DeviceRole".lower():
                    __obj = DeviceRole.objects.get(name=kwargs["filter"])
                if kwargs["model"].lower() == "Device".lower():
                    __obj = Device.objects.get(serial=kwargs["filter"])
                if kwargs["model"].lower() == "IPAddress".lower():
                    kwargs["filter"] = str(ipaddress.IPv4Network(kwargs["filter"])[0])
                    __obj = IPAddress.objects.get(address=kwargs["filter"])
                if kwargs["model"].lower() == "Site".lower():
                    __obj = Site.objects.get(name=kwargs["filter"])

                if kwargs["tag"] not in __obj.tags.all():
                    # Add Cisco DNA Center Tag to NetBox Object
                    __obj.tags.add(kwargs["tag"])
                    __obj.save()
            else:
                raise Exception("Not implemented yet")

        @staticmethod
        def site(tenant, site):
            """
            Handle Site operations with NetBox
            """

            # Gather site in Netbox (site name isn't unique, even with multiple tenants)
            if Site.objects.filter(name=site.siteNameHierarchy).exists() is False:
                Site.objects.create(
                    name=site.siteNameHierarchy,
                    slug=site.slug,
                    comments=site.id,
                    description="Managed by {}".format(tenant),
                    tenant=Tenant.objects.get(name=tenant),
                )
                sync = "Created"
            else:
                Site.objects.filter(name=site.siteNameHierarchy).update(
                    slug=site.slug,
                    comments=site.id,
                    description="Managed by {}".format(tenant),
                    tenant=Tenant.objects.get(name=tenant).id,
                )
                sync = "Updated"
            __obj = Site.objects.get(name=site.siteNameHierarchy)

            # Check if additional information is avaible for the site
            __save = False
            if len(site.additionalInfo) != 0:
                for additionalInfo in site.additionalInfo:
                    if "Location" in additionalInfo["nameSpace"]:
                        if additionalInfo["attributes"]["address"] is not None:
                            if __obj.physical_address != additionalInfo["attributes"]["address"]:
                                __obj.physical_address = additionalInfo["attributes"]["address"]
                                __save = True
                        if additionalInfo["attributes"]["latitude"] is not None:
                            if __obj.latitude != Decimal(additionalInfo["attributes"]["latitude"]):
                                __obj.latitude = additionalInfo["attributes"]["latitude"]
                                __save = True
                        if additionalInfo["attributes"]["longitude"] is not None:
                            if __obj.longitude != Decimal(additionalInfo["attributes"]["longitude"]):
                                __obj.longitude = additionalInfo["attributes"]["longitude"]
                                __save = True
            if __save is True:
                # Only update Change log if something is updated
                __obj.save()

            return Site.objects.get(name=site.siteNameHierarchy), sync

        @staticmethod
        def manufacturer(manufacture, tenant):
            """
            Handle Manufacturer operations with NetBox
            """

            # Gather manufacture in Netbox
            if Manufacturer.objects.filter(name=manufacture).exists() is False:
                Manufacturer.objects.create(
                    name=manufacture,
                    slug=manufacture.lower(),
                    description="Managed by {}".format(tenant),
                )
            else:
                Manufacturer.objects.filter(name=manufacture).update(
                    slug=manufacture.lower(),
                    description="Managed by {}".format(tenant),
                )
            return Manufacturer.objects.get(name=manufacture)

        @staticmethod
        def devicetype(manufacture, model, slug, tenant):
            """
            Handle DeviceType operations with NetBox
            """

            # Gather DeviceType in Netbox
            if DeviceType.objects.filter(manufacturer=manufacture, model=model).exists() is False:
                DeviceType.objects.create(
                    manufacturer=manufacture,
                    model=model,
                    slug=slug.lower(),
                    u_height=1,
                    comments="Managed by {}".format(tenant),
                )
            else:
                DeviceType.objects.filter(manufacturer=manufacture, model=model).update(
                    slug=slug.lower(),
                    comments="Managed by {}".format(tenant),
                )
            return DeviceType.objects.get(slug=slug.lower())

        @staticmethod
        def devicerole(role, slug, tenant):
            """
            Handle DeviceRole operations with NetBox
            """

            # Gather DeviceRole in Netbox
            if DeviceRole.objects.filter(name=role).exists() is False:
                DeviceRole.objects.create(
                    name=role,
                    slug=slug.lower(),
                    color=ColorChoices.COLOR_BLUE,
                    vm_role=False,
                    description="Managed by {}".format(tenant),
                )
            else:
                DeviceRole.objects.filter(name=role).update(
                    slug=slug.lower(),
                    color=ColorChoices.COLOR_BLUE,
                    vm_role=False,
                    description="Managed by {}".format(tenant),
                )
            return DeviceRole.objects.get(name=role)

        @staticmethod
        def device(tenant, device):
            """
            Handle Device operations with NetBox
            """

            # Check device reachability in Cisco DNA Center
            if device.reachabilityStatus == "Reachable":
                device.status = DeviceStatusChoices.STATUS_ACTIVE
            else:
                device.status = DeviceStatusChoices.STATUS_FAILED

            # Gather Device in Netbox
            if Device.objects.filter(serial=device.serialNumber).exists() is False:
                if Device.objects.filter(
                    primary_ip4=device.primary_ip4,
                    tenant=Tenant.objects.get(name=tenant).id,
                ).exists():
                    # There can't be duplicate IPs in one tenant.
                    # But DNAC can register duplicate IPs, if only one is Reachable
                    Device.objects.create(
                        name=device.hostname,
                        device_role=device.role,
                        device_type=device.family_type,
                        serial=device.serialNumber,
                        status=device.status,
                        site=device.site,
                        comments="Managed by {}".format(tenant),
                        tenant=Tenant.objects.get(name=tenant),
                    )
                    sync = "Error IPv4"
                    return Device.objects.get(serial=device.serialNumber), sync
                else:
                    Device.objects.create(
                        name=device.hostname,
                        device_role=device.role,
                        device_type=device.family_type,
                        primary_ip4=device.primary_ip4,
                        serial=device.serialNumber,
                        status=device.status,
                        site=device.site,
                        comments="Managed by {}".format(tenant),
                        tenant=Tenant.objects.get(name=tenant),
                    )
                    sync = "Created"
            else:
                if (
                    device.serialNumber
                    != Device.objects.get(
                        primary_ip4=device.primary_ip4,
                        tenant=Tenant.objects.get(name=tenant).id,
                    ).serial
                ):
                    # There can't be duplicate IPs in one tenant.
                    # But DNAC can register duplicate IPs, if only one is Reachable
                    Device.objects.filter(serial=device.serialNumber, tenant=Tenant.objects.get(name=tenant).id,).update(
                        name=device.hostname,
                        device_role=device.role,
                        device_type=device.family_type,
                        status=device.status,
                        site=device.site,
                        comments="Managed by {}".format(tenant),
                        tenant=Tenant.objects.get(name=tenant).id,
                    )
                    sync = "Error IPv4"
                else:
                    Device.objects.filter(serial=device.serialNumber, tenant=Tenant.objects.get(name=tenant).id,).update(
                        name=device.hostname,
                        device_role=device.role,
                        device_type=device.family_type,
                        primary_ip4=device.primary_ip4,
                        status=device.status,
                        site=device.site,
                        comments="Managed by {}".format(tenant),
                        tenant=Tenant.objects.get(name=tenant).id,
                    )
                    sync = "Updated"

            # Assign IP Address to Device in NetBox
            IPAddress.objects.filter(address=str(device.primary_ip4), tenant=Tenant.objects.get(name=tenant).id,).update(
                assigned_object_id=Device.objects.get(serial=device.serialNumber).id,
            )

            return Device.objects.get(serial=device.serialNumber), sync

        @staticmethod
        def ipaddress(tenant, address, hostname):
            """
            Handle IPAddress operations with NetBox
            """

            # Gather IPAddress in Netbox
            if IPAddress.objects.filter(address=address, tenant=Tenant.objects.get(name=tenant).id).exists() is False:
                IPAddress.objects.create(
                    address=address,
                    status=DeviceStatusChoices.STATUS_ACTIVE,
                    dns_name=hostname,
                    description="Managed by {}".format(tenant),
                    tenant=Tenant.objects.get(name=tenant),
                )
            else:
                IPAddress.objects.filter(address=address, tenant=Tenant.objects.get(name=tenant).id).update(
                    status=DeviceStatusChoices.STATUS_ACTIVE,
                    dns_name=hostname,
                    description="Managed by {}".format(tenant),
                    tenant=Tenant.objects.get(name=tenant).id,
                )
            return IPAddress.objects.get(address=address)

    class Purge:
        @staticmethod
        def database(**kwargs):
            """
            Purge data from NetBox Database - when running Sync
            """

            # Delete devices related to Tenant
            if kwargs["type"] == "devices":

                # Unique Serial Numbers in Netbox
                netbox_serials = [d.serial for d in Device.objects.filter(tenant=Tenant.objects.get(name=kwargs["tenant"]).id)]

                # Unique Serial Numbers in Cisco DNA Center Instance
                dnac_serials = []
                for d in kwargs["data"]:
                    dnac_serials.append(d["serial"])

                # Diff between NetBox and Cisco DNA Center Instance
                purge = list(set(netbox_serials) - set(dnac_serials))

                if len(purge) == 0:
                    return False
                else:

                    # Remove diff in NetBox
                    try:
                        for serial in purge:
                            Device.objects.filter(id=Device.objects.get(serial=serial).id).delete()
                        return True
                    except Exception as error_msg:
                        print("Error couldn't delete {}\n{}".format(kwargs["data"], error_msg))
                        return False
            # Delete sites related to Tenant
            elif kwargs["type"] == "sites":

                # Unique slug/uuid in NetBox
                netbox_sites = [s.slug for s in Site.objects.filter(tenant=Tenant.objects.get(name=kwargs["tenant"]).id)]

                # Unique Site id/uuid in Cisco DNA Center Instance
                dnac_sites = []
                for s in kwargs["data"]:
                    dnac_sites.append(s["slug"])

                # Diff between NetBox and Cisco DNA Center Instance
                purge = list(set(netbox_sites) - set(dnac_sites))

                if len(purge) == 0:
                    return False
                else:

                    # Remove diff in NetBox
                    try:
                        for uuid in purge:
                            Site.objects.filter(slug=uuid).delete()
                        return True
                    except Exception as error_msg:
                        print("Error couldn't delete {}\n{}".format(kwargs["data"], error_msg))
                        return False
            else:
                raise Exception("Not implemented yet")

        @classmethod
        def tenant(cls, **kwargs):
            """
            Delete Tenant related to Cisco DNA Center Instance
            """

            # Get NetBox tenants based on Cisco DNA Center Tag
            results = {}
            dnac_tag = System.PluginTag.get()
            tenant_name = get_object_or_404(Tenant, pk=kwargs["pk"], tags=dnac_tag).name

            # Delete objects based on dependencies until Tenant is deleted
            results[tenant_name] = {}
            results[tenant_name]["devices"] = cls.devices(**kwargs)
            results[tenant_name]["ipaddress"] = cls.ipaddress(**kwargs)
            results[tenant_name]["sites"] = cls.sites(**kwargs)
            Tenant.objects.filter(pk=kwargs["pk"]).delete()

            return results

        @classmethod
        def devices(cls, **kwargs):
            """
            Delete Devices related to Cisco DNA Center Instance
            """
            result = Device.objects.filter(tenant=kwargs["pk"]).count()
            Device.objects.filter(tenant=kwargs["pk"]).delete()
            return result

        @classmethod
        def ipaddress(cls, **kwargs):
            """
            Delete IPAddress related to Cisco DNA Center Instance
            """
            result = IPAddress.objects.filter(tenant=kwargs["pk"]).count()
            IPAddress.objects.filter(tenant=kwargs["pk"]).delete()
            return result

        @classmethod
        def sites(cls, **kwargs):
            """
            Delete Sites related to Cisco DNA Center Instance
            """
            result = Site.objects.filter(tenant=kwargs["pk"]).count()
            Site.objects.filter(tenant=kwargs["pk"]).delete()
            return result



from os import path
from setuptools import setup, find_packages


with open("requirements.txt") as f:
    requirements = f.read().splitlines()
with open(
    path.join(path.abspath(path.dirname(__file__)), "README.md"), encoding="utf-8"
) as f:
    long_description = f.read()

PACKAGE_KEYWORDS = [
    "cisco",
    "dna",
    "dnacenter",
    "python",
    "api",
    "sdk",
    "netbox",
]

setup(
    name="ciscodnacnetbox",
    version="1.0",
    description="Cisco DNA Center Integration with NetBox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/robertcsapo/ciscodnacnetbox",
    author="Robert Csapo",
    author_email="rcsapo@cisco.com",
    license="CISCO SAMPLE CODE LICENSE",
    install_requires=requirements,
    packages=find_packages(exclude=["img"]),
    package_data={"ciscodnacnetbox": ["templates/*/*.html"]},
    include_package_data=True,
    python_requires=">=3.3",
    zip_safe=False,
)



from django.db import models
from django.urls import reverse
from utilities.querysets import RestrictedQuerySet


class Settings(models.Model):
    hostname = models.CharField(max_length=2000, unique=True, blank=True, null=True)
    username = models.CharField(max_length=100)
    password = models.CharField(max_length=100)
    version = models.CharField(max_length=10)
    verify = models.BooleanField(default=False)
    status = models.BooleanField(default=True)
    objects = RestrictedQuerySet.as_manager()

    class Meta:
        app_label = "ciscodnacnetbox"
        ordering = ["hostname"]

    def __str__(self):
        return self.hostname

    def get_absolute_url(self):
        return reverse("plugins:ciscodnacnetbox:settings")



from django.urls import path
from . import views

urlpatterns = (
    # Settings
    path("settings/", views.SettingsView.as_view(), name="settings"),
    path("settings/add/", views.SettingsEdit.as_view(), name="settings_add"),
    path("settings/<int:pk>/edit/", views.SettingsEdit.as_view(), name="settings_edit"),
    path(
        "settings/<int:pk>/delete/",
        views.SettingsDelete.as_view(),
        name="settings_delete",
    ),
    path(
        "settings/delete/",
        views.SettingsDeleteBulk.as_view(),
        name="settings_delete_bulk",
    ),
    # Status
    path("status/", views.StatusView.as_view(), name="status"),
    # Tenant Data
    path("devices/", views.DeviceView.as_view(), name="devices"),
    path("<int:pk>/devices/", views.DeviceView.as_view(), name="devices"),
    path("sites/", views.SitesView.as_view(), name="sites"),
    path("<int:pk>/sites/", views.SitesView.as_view(), name="sites"),
    # Sync
    path("sync/full/", views.SyncFull.as_view(), name="sync_full"),
    path("sync/full/<uuid:id>/", views.SyncFull.as_view(), name="sync_full"),
    path(
        "sync/full/<uuid:id>/failed/",
        views.SyncFullFailed.as_view(),
        name="sync_full_failed",
    ),
    path("sync/<int:pk>/full/", views.SyncFull.as_view(), name="sync_full"),
    path("sync/sites/", views.SyncSites.as_view(), name="sync_sites"),
    path("sync/<int:pk>/sites/", views.SyncSites.as_view(), name="sync_sites"),
    path(
        "sync/devices/",
        views.SyncDevices.as_view(),
        name="sync_devices",
    ),
    path(
        "sync/<int:pk>/devices/",
        views.SyncDevices.as_view(),
        name="sync_devices",
    ),
    # Jobs
    path("job/<uuid:id>/", views.JobStatus.as_view(), name="job_status"),
    # Purge
    path("purge/<int:pk>/tenant/", views.PurgeTenant.as_view(), name="purge_tenant"),
)
